@@ -30,6 +30,7 @@
 #include <haproxy/dynbuf-t.h>
 #include <haproxy/filters-t.h>
 #include <haproxy/obj_type-t.h>
+#include <haproxy/log-t.h>
 #include <haproxy/show_flags-t.h>
 #include <haproxy/stick_table-t.h>
 #include <haproxy/vars-t.h>
@@ -155,6 +156,17 @@ enum {
 	STRM_ET_DATA_ABRT  = 0x0400,  /* data phase aborted by external cause */
 };
 
+
+/* Types of entities that may interrupt a processing, teomporarily or not
+ * depending on the context;
+ */
+enum {
+	STRM_ENTITY_NONE      = 0x0000,
+	STRM_ENTITY_RULE      = 0x0001,
+	STRM_ENTITY_FILTER    = 0x0002,
+	STRM_ENTITY_WREQ_BODY = 0x0003,
+};
+
 /* This function is used to report flags in debugging tools. Please reflect
  * below any single-bit flag addition above in the same order via the
  * __APPEND_FLAG macro. The new end of the buffer is returned.
@@ -234,6 +246,9 @@ struct stream {
 	int conn_retries;               /* number of connect retries performed */
 	unsigned int conn_exp;          /* wake up time for connect, queue, turn-around, ... */
 	unsigned int conn_err_type;     /* first error detected, one of STRM_ET_* */
+
+	struct stream *parent;          /* Pointer to the parent stream, if any. NULL most of time */
+
 	struct list list;               /* position in the thread's streams list */
 	struct mt_list by_srv;          /* position in server stream list */
 	struct list back_refs;          /* list of users tracking this stream */
@@ -242,9 +257,13 @@ struct stream {
 	uint64_t lat_time;		/* total latency time experienced */
 	uint64_t cpu_time;              /* total CPU time consumed */
 	struct freq_ctr call_rate;      /* stream task call rate without making progress */
+	uint32_t passes_stconn;         /* number of passes on the stconn evaluation code */
+	uint32_t passes_reqana;         /* number of passes on the req analysers block */
+	uint32_t passes_resana;         /* number of passes on the res analysers block */
+	uint32_t passes_propag;         /* number of passes on the shut/err propag code */
 
+	unsigned short max_retries;     /* Maximum number of connection retried (=0 is backend is not set) */
 	short store_count;
-	/* 2 unused bytes here */
 
 	struct {
 		struct stksess *ts;
@@ -265,7 +284,8 @@ struct stream {
 
 	struct strm_logs logs;                  /* logs for this stream */
 
-	void (*do_log)(struct stream *s);       /* the function to call in order to log (or NULL) */
+	void (*do_log)(struct stream *s,        /* the function to call in order to log (or NULL) */
+	               struct log_orig origin);
 	void (*srv_error)(struct stream *s,     /* the function to call upon unrecoverable server errors (or NULL) */
 			  struct stconn *sc);
 
@@ -280,8 +300,16 @@ struct stream {
 	void *current_rule;                     /* this is used to store the current rule to be resumed. */
 	int rules_exp;                          /* expiration date for current rules execution */
 	int tunnel_timeout;
-	const char *last_rule_file;             /* last evaluated final rule's file (def: NULL) */
-	int last_rule_line;                     /* last evaluated final rule's line (def: 0) */
+
+	struct {
+		void *ptr;                      /* Pointer on the entity  (def: NULL) */
+		int type;                       /* entity type (STRM_ENTITY_*) */
+	} last_entity;                          /* last evaluated entity that interrupted processing */
+
+	struct {
+		void *ptr;                      /* Pointer on the entity  (def: NULL) */
+		int type;                       /* entity type (STRM_ENTITY_*) */
+	} waiting_entity;                       /* The entity waiting to continue its processing and interrupted by an error/timeout */
 
 	unsigned int stream_epoch;              /* copy of stream_epoch when the stream was created */
 	struct hlua *hlua[2];                   /* lua runtime context (0: global, 1: per-thread) */
