@@ -51,11 +51,6 @@
 
 /* per-thread quic datagram handlers */
 struct quic_dghdlr *quic_dghdlrs;
-struct eb_root *quic_cid_tree;
-
-/* global CID trees */
-#define QUIC_CID_TREES_CNT 256
-struct quic_cid_tree *quic_cid_trees;
 
 /* Size of the internal buffer of QUIC RX buffer at the fd level */
 #define QUIC_RX_BUFSZ  (1UL << 18)
@@ -69,6 +64,7 @@ static void quic_disable_listener(struct listener *listener);
 static int quic_bind_tid_prep(struct connection *conn, int new_tid);
 static void quic_bind_tid_commit(struct connection *conn);
 static void quic_bind_tid_reset(struct connection *conn);
+static int quic_get_info(struct connection *conn, long long int *info, int info_num);
 
 /* Note: must not be declared <const> as its list will be overwritten */
 struct protocol proto_quic4 = {
@@ -87,6 +83,7 @@ struct protocol proto_quic4 = {
 	.get_src        = quic_sock_get_src,
 	.get_dst        = quic_sock_get_dst,
 	.connect        = quic_connect_server,
+	.get_info       = quic_get_info,
 	.bind_tid_prep   = quic_bind_tid_prep,
 	.bind_tid_commit = quic_bind_tid_commit,
 	.bind_tid_reset  = quic_bind_tid_reset,
@@ -107,8 +104,6 @@ struct protocol proto_quic4 = {
 	.rx_unbind      = sock_unbind,
 	.rx_listening   = quic_sock_accepting_conn,
 	.default_iocb   = quic_lstnr_sock_fd_iocb,
-	.receivers      = LIST_HEAD_INIT(proto_quic4.receivers),
-	.nb_receivers   = 0,
 #ifdef SO_REUSEPORT
 	.flags          = PROTO_F_REUSEPORT_SUPPORTED,
 #endif
@@ -133,6 +128,7 @@ struct protocol proto_quic6 = {
 	.get_src        = quic_sock_get_src,
 	.get_dst        = quic_sock_get_dst,
 	.connect        = quic_connect_server,
+	.get_info       = quic_get_info,
 	.bind_tid_prep   = quic_bind_tid_prep,
 	.bind_tid_commit = quic_bind_tid_commit,
 	.bind_tid_reset  = quic_bind_tid_reset,
@@ -153,8 +149,6 @@ struct protocol proto_quic6 = {
 	.rx_unbind      = sock_unbind,
 	.rx_listening   = quic_sock_accepting_conn,
 	.default_iocb   = quic_lstnr_sock_fd_iocb,
-	.receivers      = LIST_HEAD_INIT(proto_quic6.receivers),
-	.nb_receivers   = 0,
 #ifdef SO_REUSEPORT
 	.flags          = PROTO_F_REUSEPORT_SUPPORTED,
 #endif
@@ -522,47 +516,73 @@ static int quic_alloc_rxbufs_listener(struct listener *l)
 	return 0;
 }
 
-/* Check if platform supports the required feature set for quic-conn owned
- * socket. <l> listener must already be binded; a dummy socket will be opened
- * on the same address as one of the support test.
- *
- * Returns true if platform is deemed compatible else false.
+/* Check for platform support of a set of advanced UDP network API features
+ * used by haproxy QUIC stack. Automatically disable unsupported features.
+ * Listener <l> serves to test the ability of binding multiple sockets on the
+ * same address.
  */
-static int quic_test_sock_per_conn_support(struct listener *l)
+static int quic_test_socketopts(struct listener *l)
 {
 	const struct receiver *rx = &l->rx;
-	int ret = 1, fdtest;
+	int fdtest = -1;
 
 	/* Check if IP destination address can be retrieved on recvfrom()
 	 * operation.
 	 */
-#if !defined(IP_PKTINFO) && !defined(IP_RECVDSTADDR)
-	ha_diag_warning("Your platform does not seem to support UDP source address retrieval through IP_PKTINFO or an alternative flag. "
-			"QUIC connections will use listener socket.\n");
-	ret = 0;
-#endif
-
-	/* Check if platform support multiple UDP sockets bind on the same
-	 * local address. Create a dummy socket and bind it on the same address
-	 * as <l> listener. If bind system call fails, deactivate socket per
-	 * connection. All other errors are not taken into account.
-	 */
-	if (ret) {
+	if (global.tune.options & GTUNE_QUIC_SOCK_PER_CONN) {
 		fdtest = socket(rx->proto->fam->sock_domain,
-		                rx->proto->sock_type, rx->proto->sock_prot);
-		if (fdtest >= 0) {
-			if (setsockopt(fdtest, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) &&
-			    bind(fdtest, (struct sockaddr *)&rx->addr, rx->proto->fam->sock_addrlen) < 0) {
-				ha_diag_warning("Your platform does not seem to support multiple UDP sockets binded on the same address. "
-						"QUIC connections will use listener socket.\n");
-				ret = 0;
-			}
+				rx->proto->sock_type, rx->proto->sock_prot);
+		if (fdtest < 0)
+			goto err;
 
-			close(fdtest);
+#if defined(IP_PKTINFO) || defined(IP_RECVDSTADDR)
+		/* Check if platform support multiple UDP sockets bind on the same
+		 * local address. Create a dummy socket and bind it on the same address
+		 * as <l> listener. If bind system call fails, deactivate socket per
+		 * connection. All other errors are not taken into account.
+		 */
+		if (setsockopt(fdtest, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) &&
+		    bind(fdtest, (struct sockaddr *)&rx->addr, rx->proto->fam->sock_addrlen) < 0) {
+			ha_diag_warning("Your platform does not seem to support multiple UDP sockets binded on the same address. "
+			                "QUIC connections will use listener socket.\n");
+			global.tune.options &= ~GTUNE_QUIC_SOCK_PER_CONN;
 		}
+#else
+		ha_diag_warning("Your platform does not seem to support UDP source address retrieval through IP_PKTINFO or an alternative flag. "
+		                "QUIC connections will use listener socket.\n");
+		global.tune.options &= ~GTUNE_QUIC_SOCK_PER_CONN;
+#endif
 	}
 
-	return ret;
+	/* Check for UDP GSO support. */
+	if (!(global.tune.options & GTUNE_QUIC_NO_UDP_GSO)) {
+		if (fdtest < 0) {
+			fdtest = socket(rx->proto->fam->sock_domain,
+					rx->proto->sock_type, rx->proto->sock_prot);
+			if (fdtest < 0)
+				goto err;
+		}
+
+#ifdef UDP_SEGMENT
+		if (setsockopt(fdtest, SOL_UDP, UDP_SEGMENT, &zero, sizeof(zero))) {
+			ha_diag_warning("Your platform does not support UDP GSO. "
+			                "This will be automatically disabled for QUIC transfer.\n");
+			global.tune.options |= GTUNE_QUIC_NO_UDP_GSO;
+		}
+#else
+		ha_diag_warning("Your platform does not support UDP GSO. "
+		                "This will be automatically disabled for QUIC transfer.\n");
+		global.tune.options |= GTUNE_QUIC_NO_UDP_GSO;
+#endif
+	}
+
+	if (fdtest >= 0)
+		close(fdtest);
+	return ERR_NONE;
+
+ err:
+	ha_alert("Fatal error on quic_test_sockopts(): %s.\n", strerror(errno));
+	return ERR_FATAL;
 }
 
 /* This function tries to bind a QUIC4/6 listener. It may return a warning or
@@ -626,10 +646,8 @@ static int quic_bind_listener(struct listener *listener, char *errmsg, int errle
 		goto udp_return;
 	}
 
-	if (global.tune.options & GTUNE_QUIC_SOCK_PER_CONN) {
-		if (!quic_test_sock_per_conn_support(listener))
-			global.tune.options &= ~GTUNE_QUIC_SOCK_PER_CONN;
-	}
+	if (quic_test_socketopts(listener))
+		return ERR_FATAL;
 
 	if (global.tune.frontend_rcvbuf)
 		setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &global.tune.frontend_rcvbuf, sizeof(global.tune.frontend_rcvbuf));
@@ -675,6 +693,21 @@ static void quic_disable_listener(struct listener *l)
 		fd_stop_recv(l->rx.fd);
 }
 
+static int quic_get_info(struct connection *conn, long long int *info, int info_num)
+{
+	struct quic_conn *qc = conn->handle.qc;
+
+	switch (info_num) {
+	case 0:  *info = qc->path->loss.srtt * 1000;      break;
+	case 1:  *info = qc->path->loss.rtt_var * 1000;   break;
+	case 4:  *info = qc->path->loss.nb_lost_pkt;      break;
+	case 7:  *info = qc->path->loss.nb_reordered_pkt; break;
+	default: return 0;
+	}
+
+	return 1;
+}
+
 /* change the connection's thread to <new_tid>. For frontend connections, the
  * target is a listener, and the caller is responsible for guaranteeing that
  * the listener assigned to the connection is bound to the requested thread.
@@ -723,17 +756,6 @@ static int quic_alloc_dghdlrs(void)
 		MT_LIST_INIT(&dghdlr->dgrams);
 	}
 
-	quic_cid_trees = calloc(QUIC_CID_TREES_CNT, sizeof(*quic_cid_trees));
-	if (!quic_cid_trees) {
-		ha_alert("Failed to allocate global quic CIDs trees.\n");
-		return 0;
-	}
-
-	for (i = 0; i < QUIC_CID_TREES_CNT; ++i) {
-		HA_RWLOCK_INIT(&quic_cid_trees[i].lock);
-		quic_cid_trees[i].root = EB_ROOT_UNIQUE;
-	}
-
 	return 1;
 }
 REGISTER_POST_CHECK(quic_alloc_dghdlrs);
@@ -748,8 +770,6 @@ static int quic_deallocate_dghdlrs(void)
 		free(quic_dghdlrs);
 	}
 
-	ha_free(&quic_cid_trees);
-
 	return 1;
 }
 REGISTER_POST_DEINIT(quic_deallocate_dghdlrs);
