@@ -519,18 +519,28 @@ int quic_build_post_handshake_frames(struct quic_conn *qc)
 			goto err;
 		}
 
-		conn_id = new_quic_cid(qc->cids, qc, NULL, NULL);
+		conn_id = quic_cid_alloc(qc);
 		if (!conn_id) {
 			qc_frm_free(qc, &frm);
 			TRACE_ERROR("CID allocation error", QUIC_EV_CONN_IO_CB, qc);
 			goto err;
 		}
 
-		/* TODO To prevent CID tree locking, all CIDs created here
-		 * could be allocated at the same time as the first one.
-		 */
-		_quic_cid_insert(conn_id);
+		if (quic_cid_generate(conn_id)) {
+			qc_frm_free(qc, &frm);
+			pool_free(pool_head_quic_connection_id, conn_id);
+			TRACE_ERROR("error on CID generation", QUIC_EV_CONN_IO_CB, qc);
+			goto err;
+		}
 
+		if (quic_cid_insert(conn_id, NULL)) {
+			TRACE_STATE("collision during CID generation on post-handshake", QUIC_EV_CONN_IO_CB, qc);
+			pool_free(pool_head_quic_connection_id, conn_id);
+			qc_frm_free(qc, &frm);
+			continue;
+		}
+
+		quic_cid_register_seq_num(conn_id);
 		quic_connection_id_to_frm_cpy(frm, conn_id);
 		LIST_APPEND(&frm_list, &frm->list);
 	}
@@ -1117,6 +1127,14 @@ struct quic_conn *qc_new_conn(const struct quic_version *qv, int ipv4,
 	/* Required to safely call quic_conn_prx_cntrs_update() from quic_conn_release(). */
 	qc->prx_counters = NULL;
 
+	qc->cids = pool_alloc(pool_head_quic_cids);
+	if (!qc->cids) {
+		TRACE_ERROR("Could not allocate a new CID tree", QUIC_EV_CONN_INIT, qc);
+		goto err;
+	}
+	*qc->cids = EB_ROOT;
+	qc->next_cid_seq_num = 0;
+
 	/* QUIC Server (or listener). */
 	if (server) {
 		cc_algo = l->bind_conf->quic_cc_algo;
@@ -1165,16 +1183,7 @@ struct quic_conn *qc_new_conn(const struct quic_version *qv, int ipv4,
 		goto err;
 	}
 
-	qc->cids = pool_alloc(pool_head_quic_cids);
-	if (!qc->cids) {
-		TRACE_ERROR("Could not allocate a new CID tree", QUIC_EV_CONN_INIT, qc);
-		goto err;
-	}
-	*qc->cids = EB_ROOT;
-
-	conn_id->qc = qc;
-
-	if (HA_ATOMIC_LOAD(&l->rx.quic_mode) == QUIC_SOCK_MODE_CONN &&
+	if (l && HA_ATOMIC_LOAD(&l->rx.quic_mode) == QUIC_SOCK_MODE_CONN &&
 	    (global.tune.options & GTUNE_QUIC_SOCK_PER_CONN) &&
 	    is_addr(local_addr)) {
 		TRACE_USER("Allocate a socket for QUIC connection", QUIC_EV_CONN_INIT, qc);
@@ -1553,9 +1562,13 @@ void qc_idle_timer_do_rearm(struct quic_conn *qc, int arm_ack)
 {
 	unsigned int expire;
 
+	TRACE_ENTER(QUIC_EV_CONN_IDLE_TIMER, qc);
+
 	/* It is possible the idle timer task has been already released. */
-	if (!qc->idle_timer_task)
-		return;
+	if (!qc->idle_timer_task) {
+		TRACE_PROTO("idle timer already released", QUIC_EV_CONN_IDLE_TIMER, qc);
+		goto leave;
+	}
 
 	if (qc->flags & (QUIC_FL_CONN_CLOSING|QUIC_FL_CONN_DRAINING)) {
 		/* RFC 9000 10.2. Immediate Close
@@ -1584,6 +1597,8 @@ void qc_idle_timer_do_rearm(struct quic_conn *qc, int arm_ack)
 		expire = QUIC_MAX(3 * quic_pto(qc), qc->max_idle_timeout);
 	}
 
+	TRACE_PRINTF(TRACE_LEVEL_PROTO, QUIC_EV_CONN_IDLE_TIMER, qc, 0, 0, 0,
+	             "quic_pto=%u expire=%u", quic_pto(qc), expire);
 	qc->idle_expire = tick_add(now_ms, MS_TO_TICKS(expire));
 	/* Note that the ACK timer is not armed during the handshake. So,
 	 * the handshake expiration date is taken into an account only
@@ -1605,6 +1620,9 @@ void qc_idle_timer_do_rearm(struct quic_conn *qc, int arm_ack)
 		task_queue(qc->idle_timer_task);
 		TRACE_PROTO("idle timer armed", QUIC_EV_CONN_IDLE_TIMER, qc);
 	}
+
+ leave:
+	TRACE_LEAVE(QUIC_EV_CONN_IDLE_TIMER, qc);
 }
 
 /* Rearm the idle timer or ack timer for <qc> QUIC connection depending on <read>
