@@ -684,7 +684,8 @@ static void sc_app_shut(struct stconn *sc)
 	}
 
 	sc->flags &= ~SC_FL_NOLINGER;
-	sc->flags |= SC_FL_ABRT_DONE;
+	if (!(sc->flags & (SC_FL_EOS|SC_FL_ABRT_DONE)))
+		sc->flags |= SC_FL_ABRT_DONE;
 	if (sc->flags & SC_FL_ISBACK)
 		__sc_strm(sc)->conn_exp = TICK_ETERNITY;
 
@@ -819,7 +820,8 @@ static void sc_app_shut_conn(struct stconn *sc)
 	}
 
 	sc->flags &= ~SC_FL_NOLINGER;
-	sc->flags |= SC_FL_ABRT_DONE;
+	if (!(sc->flags & (SC_FL_EOS|SC_FL_ABRT_DONE)))
+		sc->flags |= SC_FL_ABRT_DONE;
 	if (sc->flags & SC_FL_ISBACK)
 		__sc_strm(sc)->conn_exp = TICK_ETERNITY;
 }
@@ -1002,7 +1004,8 @@ static void sc_app_shut_applet(struct stconn *sc)
 	}
 
 	sc->flags &= ~SC_FL_NOLINGER;
-	sc->flags |= SC_FL_ABRT_DONE;
+	if (!(sc->flags & (SC_FL_EOS|SC_FL_ABRT_DONE)))
+		sc->flags |= SC_FL_ABRT_DONE;
 	if (sc->flags & SC_FL_ISBACK)
 		__sc_strm(sc)->conn_exp = TICK_ETERNITY;
 }
